# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
import mmap


class Rangeable:
    """
    Interface for all rangeable objects.

    All rangeable objects must be able to return their base and max
    addresses.
    """
    def get_base_address(self):
        """
        Getter for the base address of the object.

        :return: The base address of the object.
        :rtype: int
        """
        raise NotImplementedError

    def get_max_address(self):
        """
        Getter for the max address of the object.

        :return: The max address of the object.
        :rtype: int
        """
        raise NotImplementedError


class StackTrace:
    def __init__(self, trace=None):
        self.trace = trace

    def __str__(self):
        ret = ""
        if self.trace is not None:
            for line in self.trace:
                ret += "    by\t{}\n".format(line)
        return ret


def memory_map(filename, size=0, access=mmap.ACCESS_WRITE, offset=0):
    """
    Memory map a file.

    :Warning:
        `offset` has to be a non-negative multiple of PAGESIZE or
        ALLOCATIONGRANULARITY

    :param filename: The file to be mapped.
    :type filename: str
    :param size: Number of bytes to be mapped. If is equal 0, the
        whole file at the moment of the call will be mapped.
    :type size: int
    :param offset: The offset within the file to be mapped.
    :type offset: int
    :param access: The type of access provided to mmap.
    :return: The mapped file.
    :rtype: mmap.mmap
    """
    fd = os.open(filename, os.O_RDWR)
    m_file = mmap.mmap(fd, size, access=access, offset=offset)
    os.close(fd)
    return m_file


def range_cmp(lhs, rhs):
    """
    A range compare function.

    :param lhs: The left hand side of the comparison.
    :type lhs: Rangeable
    :param rhs: The right hand side of the comparison.
    :type rhs: Rangeable
    :return: -1 if lhs is before rhs, 1 when after and 0 on overlap.
    :rtype: int

    The comparison function may be explained as::

        Will return -1:
        |___lhs___|
                        |___rhs___|

        Will return +1:
        |___rhs___|
                        |___lhs___|

        Will return 0:
        |___lhs___|
               |___rhs___|
    """
    if lhs.get_max_address() <= rhs.get_base_address():
        return -1
    elif lhs.get_base_address() >= rhs.get_max_address():
        return 1
    else:
        return 0



#!/usr/bin/python
#
# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import argparse
import statemachine
import opscontext
import consistencycheckwrap
import loggingfacility
import markerparser
import sys
import reorderengines


def main():
    # TODO unicode support
    # TODO parameterize reorder engine type
    parser = argparse.ArgumentParser(description="Store reordering tool")
    parser.add_argument("-l", "--logfile",
                        required=True,
                        help="the pmemcheck log file to process")
    parser.add_argument("-c", "--checker",
                        choices=consistencycheckwrap.checkers,
                        default=consistencycheckwrap.checkers[0],
                        help="choose consistency checker type")
    parser.add_argument("-p", "--path",
                        required=True,
                        help="path to the consistency checker and arguments",
                        nargs='+')
    parser.add_argument("-n", "--name",
                        help="consistency check function " +
                        "for the 'lib' checker")
    parser.add_argument("-o", "--output",
                        help="set the logger output file")
    parser.add_argument("-e", "--output-level",
                        choices=loggingfacility.log_levels,
                        help="set the output log level")
    parser.add_argument("-x", "--extended-macros",
                        help="list of pairs MARKER=ENGINE or " +
                        "json config file")
    engines_keys = list(reorderengines.engines.keys())
    parser.add_argument("-r", "--default-engine",
                        help="set default reorder engine " +
                        "default=NoReorderNoChecker",
                        choices=engines_keys,
                        default=engines_keys[0])
    args = parser.parse_args()
    logger = loggingfacility.get_logger(
                                        args.output,
                                        args.output_level)
    checker = consistencycheckwrap.get_checker(
                                               args.checker,
                                               ' '.join(args.path),
                                               args.name)

    markers = markerparser.MarkerParser().get_markers(args.extended_macros)

    # create the script context
    context = opscontext.OpsContext(
                                    args.logfile,
                                    checker,
                                    logger,
                                    args.default_engine,
                                    markers)

    # init and run the state machine
    a = statemachine.StateMachine(statemachine.InitState(context))
    if a.run_all(context.extract_operations()) is False:
        sys.exit(1)


if __name__ == "__main__":
    main()



#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""Helper variables, functions and classes"""

from os.path import join, abspath, dirname
import sys


# Constant paths to repository elements
ROOTDIR = abspath(join(dirname(__file__), '..'))

WIN_DEBUG_BUILDDIR = abspath(join(ROOTDIR, '..', 'x64', 'Debug'))
WIN_DEBUG_EXEDIR = abspath(join(WIN_DEBUG_BUILDDIR, 'tests'))

WIN_NONDEBUG_BUILDDIR = abspath(join(ROOTDIR, '..', 'x64', 'Release'))
WIN_NONDEBUG_EXEDIR = abspath(join(WIN_NONDEBUG_BUILDDIR, 'tests'))

if sys.platform == 'win32':
    DEBUG_LIBDIR = abspath(join(WIN_DEBUG_BUILDDIR, 'libs'))
    NONDEBUG_LIBDIR = abspath(join(WIN_NONDEBUG_BUILDDIR, 'libs'))
else:
    DEBUG_LIBDIR = abspath(join(ROOTDIR, '..', 'debug'))
    NONDEBUG_LIBDIR = abspath(join(ROOTDIR, '..', 'nondebug'))

#
# KiB, MiB, GiB ... -- byte unit prefixes
#

KiB = 2 ** 10
MiB = 2 ** 20
GiB = 2 ** 30
TiB = 2 ** 40
PiB = 2 ** 50


class Color:
    """
    Set the font color. This functionality relies on ANSI espace sequences
    and is currently disabled for Windows
    """
    if sys.platform != 'win32':
        RED = '\33[91m'
        GREEN = '\33[92m'
        END = '\33[0m'
    else:
        RED, GREEN, END = '', '', ''


class Message:
    """Simple level based logger"""
    def __init__(self, config):
        self.config = config

    def print(self, msg):
        if self.config.unittest_log_level >= 1:
            print(msg)

    def print_verbose(self, msg):
        if self.config.unittest_log_level >= 2:
            print(msg)


def filter_contexts(config_ctx, test_ctx):
    """
    Return contexts that should be used in execution based on
    contexts provided by config and test case
    """
    if not test_ctx:
        return [c for c in config_ctx if not c.explicit]
    return [c for c in config_ctx if c in test_ctx]


def run_tests_common(testcases, config):
    """
    Common implementation for running tests - used by RUNTESTS.py and
    single test case interpreter
    """
    if config.test_sequence:
        # filter test cases from sequence
        testcases = [t for t in testcases if t.testnum in config.test_sequence]
        # sort testcases so their sequence matches provided test sequence
        testcases.sort(key=lambda tc: config.test_sequence.index(tc.testnum))

    if not testcases:
        sys.exit('No testcases to run found for selected configuration.')

    results = [t(config)._execute() for t in testcases]
    if any(results):  # if any test failed
        return 1
    return 0



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from operationfactory import OperationFactory
from binaryoutputhandler import BinaryOutputHandler
import reorderengines
import memoryoperations
from itertools import repeat


class OpsContext:
    """
    Holds the context of the performed operations.

    :ivar _operations: The operations to be performed, based on the log file.
    :type _operations: list of strings
    :ivar reorder_engine: The reordering engine used at the moment.
    :type one of the reorderengine Class
    :ivar default_engine: The default reordering engine.
    :type default_engine: One of the reorderengines Class
    :ivar test_on_barrier: Check consistency on barrier.
    :type test_on_barrier: bool
    :ivar default_barrier: Default consistency barrier status.
    :type default_barrier: bool
    :ivar file_handler: The file handler used.
    """
    def __init__(self, log_file, checker, logger, arg_engine, markers):
        """
        Splits the operations in the log file and sets the instance variables
        to default values.

        :param log_file: The full name of the log file.
        :type log_file: str
        :return: None
        """
        # TODO reading the whole file at once is rather naive
        # change in the future
        self._operations = open(log_file).read().split("|")
        engine = reorderengines.get_engine(arg_engine)
        self.reorder_engine = engine
        self.test_on_barrier = engine.test_on_barrier
        self.default_engine = self.reorder_engine
        self.default_barrier = self.default_engine.test_on_barrier
        self.file_handler = BinaryOutputHandler(checker)
        self.checker = checker
        self.logger = logger
        self.markers = markers
        self.stack_engines = [('START', getattr(memoryoperations, arg_engine))]

    # TODO this should probably be made a generator
    def extract_operations(self):
        """
        Creates specific operation objects based on the labels available
        in the split log file.

        :return: list of subclasses of :class:`memoryoperations.BaseOperation`
        """
        stop_index = start_index = 0

        for i, elem in enumerate(self._operations):
            if "START" in elem:
                start_index = i
            elif "STOP" in elem:
                stop_index = i

        return list(map(OperationFactory.create_operation,
                        self._operations[start_index + 1:stop_index],
                        repeat(self.markers), repeat(self.stack_engines)))



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import memoryoperations
from reorderexceptions import NotSupportedOperationException


class OperationFactory:
    """
    An abstract memory operation factory.

    This object factory puts special constraints on names of classes.
    It creates objects based on log in string format, as such the
    classes have to start with a capital letter and the rest of the
    name has to be in lowercase. For example::

        STORE -> Store
        FULL_REORDER -> Full_reorder

    The object to be created has to have and internal **Factory** class
    with a :func:`create` method taking a string parameter. For example see
    :class:`memoryoperations.Store`.

    :cvar __factories: The registered object factories.
    :type __factories: dict
    """
    __factories = {}
    __suffix = ['.BEGIN', '.END']
    memoryoperations.BaseOperation()

    @staticmethod
    def add_factory(id_, operation_factory):
        """
        Explicitly register an object factory.

        This method should be used when the factory cannot be inferred
        from the name of the object to be created.

        :param id_: The id under which this factory is to be registered
            in the dictionary.
        :type id_: str
        :param operation_factory: The operation factory to be registered.
        :return: None
        """
        OperationFactory.__factories[id_] = operation_factory

    @staticmethod
    def create_operation(string_operation, markers, stack):

        def check_marker_format(marker):
            """
            Checks if marker has proper suffix.
            """
            for s in OperationFactory.__suffix:
                if marker.endswith(s):
                    return

            raise NotSupportedOperationException(
                        "Incorrect marker format {}, suffix is missing."
                        .format(marker))

        def check_pair_consistency(stack, marker):
            """
            Checks if markers do not cross.
            You can pop from stack only if end
            marker match previous one.

            Example OK:
                MACRO1.BEGIN
                    MACRO2.BEGIN
                    MACRO2.END
                MACRO1.END

            Example NOT OK:
                MACRO1.BEGIN
                    MACRO2.BEGIN
                MACRO1.END
                    MACRO2.END
            """
            top = stack[-1][0]
            if top.endswith(OperationFactory.__suffix[0]):
                    top = top[:-len(OperationFactory.__suffix[0])]
            if marker.endswith(OperationFactory.__suffix[-1]):
                    marker = marker[:-len(OperationFactory.__suffix[-1])]

            if top != marker:
                raise NotSupportedOperationException(
                        "Cannot cross markers: {0}, {1}"
                        .format(top, marker))

        """
        Creates the object based on the pre-formatted string.

        The string needs to be in the specific format. Each specific value
        in the string has to be separated with a `;`. The first field
        has to be the name of the operation, the rest are operation
        specific values.

        :param string_operation: The string describing the operation.
        :param markers: The dict describing the pair marker-engine.
        :param stack: The stack describing the order of engine changes.
        :return: The specific object instantiated based on the string.
        """
        id_ = string_operation.split(";")[0]
        id_case_sensitive = id_.lower().capitalize()

        # checks if id_ is one of memoryoperation classes
        mem_ops = getattr(memoryoperations, id_case_sensitive, None)

        # if class is not one of memoryoperations
        # it means it can be user defined marker
        if mem_ops is None:
            check_marker_format(id_)
            # if id_ is section BEGIN
            if id_.endswith(OperationFactory.__suffix[0]):
                # BEGIN defined by user
                marker_name = id_.partition('.')[0]
                if markers is not None and marker_name in markers:
                    engine = markers[marker_name]
                    try:
                        mem_ops = getattr(memoryoperations, engine)
                    except AttributeError:
                        raise NotSupportedOperationException(
                                "Not supported reorder engine: {}"
                                .format(engine))
                # BEGIN but not defined by user
                else:
                    mem_ops = stack[-1][1]

                if issubclass(mem_ops, memoryoperations.ReorderBase):
                    stack.append((id_, mem_ops))

            # END section
            elif id_.endswith(OperationFactory.__suffix[-1]):
                check_pair_consistency(stack, id_)
                stack.pop()
                mem_ops = stack[-1][1]

        # here we have proper memory operation to perform,
        # it can be Store, Fence, ReorderDefault etc.
        id_ = mem_ops.__name__
        if id_ not in OperationFactory.__factories:
            OperationFactory.__factories[id_] = mem_ops.Factory()

        return OperationFactory.__factories[id_].create(string_operation)



#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Set of classes that represent the context of single test execution
(like build, filesystem, test type)"""

import os
import sys
import itertools
import subprocess as sp

import helpers as hlp


def expand(*classes):
    """Return flatten list of container classes with removed duplications"""
    return list(set(itertools.chain(*classes)))


class Context:
    """Manage test execution based on values from context classes"""

    def __init__(self, test, conf, fs, build):
        self.env = {}
        for ctx in [fs, build]:
            if hasattr(ctx, 'env'):
                self.env.update(ctx.env)
        self.test = test
        self.conf = conf
        self.build = build
        self.msg = hlp.Message(conf)
        self.fs = fs

    @property
    def testdir(self):
        """Test directory on selected filesystem"""
        # Testdir uses 'fs.dir' field  - it is illegal to access it in case of
        # 'Non' fs. Hence it is implemented as a property.
        return os.path.join(self.fs.dir, self.test.testdir)

    def create_holey_file(self, size, name):
        """Create a new file with the selected size and name"""
        filepath = os.path.join(self.testdir, name)
        with open(filepath, 'w') as f:
            f.seek(size)
            f.write('\0')
        return filepath

    def exec(self, cmd, args='', expected_exit=0):
        """Execute binary in current test context"""
        env = {**self.env, **os.environ.copy(), **self.test.utenv}
        if sys.platform == 'win32':
            env['PATH'] = self.build.libdir + os.pathsep + env.get('PATH', '')
            cmd = os.path.join(self.build.exedir, cmd) + '.exe'
        else:
            env['LD_LIBRARY_PATH'] = self.build.libdir + os.pathsep +\
                                     env.get('LD_LIBRARY_PATH', '')
            cmd = os.path.join(self.test.cwd, cmd) + self.build.exesuffix

        proc = sp.run([cmd, args], env=env, cwd=self.test.cwd,
                      timeout=self.conf.timeout, stdout=sp.PIPE,
                      stderr=sp.STDOUT, universal_newlines=True)

        if proc.returncode != expected_exit:
            self.test.fail(proc.stdout)
        else:
            self.msg.print_verbose(proc.stdout)


class _CtxType(type):
    """Metaclass for context classes that can stand for multiple classes"""
    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        # it is supposed that context class has a user defined base class
        if cls.__base__.__name__ == 'object':
            return

        # context class preferred to be used when 'Any' test option is provided
        if not hasattr(cls, 'preferred'):
            cls.is_preferred = False

        # if explicit, run only if the test case explicitly specifies this
        # context in its own configuration
        if not hasattr(cls, 'explicit'):
            cls.explicit = False

        if hasattr(cls, 'includes'):
            cls.includes = expand(*cls.includes)
        else:
            # class that does not include other classes, includes itself
            cls.includes = [cls]

        # set instance (not class) method
        setattr(cls, '__repr__', lambda cls: cls.__class__.__name__.lower())

    def __repr__(cls):
        return cls.__name__.lower()

    def __iter__(cls):
        for c in cls.includes:
            yield c

    def factory(cls, conf, *classes):
        return [c(conf) for c in expand(*classes)]


class _Build(metaclass=_CtxType):
    """Base and factory class for standard build classes"""
    exesuffix = ''


class Debug(_Build):
    """Set the context for debug build"""
    is_preferred = True

    def __init__(self, conf):
        if sys.platform == 'win32':
            self.exedir = hlp.WIN_DEBUG_EXEDIR
        self.libdir = hlp.DEBUG_LIBDIR


class Nondebug(_Build):
    """Set the context for nondebug build"""
    is_preferred = True

    def __init__(self, conf):
        if sys.platform == 'win32':
            self.exedir = hlp.WIN_NONDEBUG_EXEDIR
        self.libdir = hlp.NONDEBUG_LIBDIR


# Build types not available on Windows
if sys.platform != 'win32':
    class Static_Debug(_Build):
        """Sets the context for static_debug build"""

        def __init__(self, conf):
            self.exesuffix = '.static-debug'
            self.libdir = hlp.DEBUG_LIBDIR

    class Static_Nondebug(_Build):
        """Sets the context for static_nondebug build"""

        def __init__(self, conf):
            self.exesuffix = '.static-nondebug'
            self.libdir = hlp.NONDEBUG_LIBDIR


class _Fs(metaclass=_CtxType):
    """Base class for filesystem classes"""


class Pmem(_Fs):
    """Set the context for pmem filesystem"""
    is_preferred = True

    def __init__(self, conf):
        self.dir = os.path.abspath(conf.pmem_fs_dir)
        if conf.fs_dir_force_pmem == 1:
            self.env = {'PMEM_IS_PMEM_FORCE': '1'}


class Nonpmem(_Fs):
    """Set the context for nonpmem filesystem"""
    def __init__(self, conf):
        self.dir = os.path.abspath(conf.non_pmem_fs_dir)


class Non(_Fs):
    """
    No filesystem is used. Accessing some fields of this class is prohibited.
    """
    explicit = True

    def __init__(self, conf):
        pass

    def __getattribute__(self, name):
        if name in ('dir',):
            raise AttributeError("fs '{}' attribute cannot be used for '{}' fs"
                                 .format(name, self))
        else:
            return super().__getattribute__(name)


class _TestType(metaclass=_CtxType):
    """Base class for test duration"""


class Short(_TestType):
    pass


class Medium(_TestType):
    pass


class Long(_TestType):
    pass


class Check(_TestType):
    includes = [Short, Medium]



# Copyright 2018, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from sys import exit
from os import path
from ctypes import cdll, c_char_p, c_int
import os

checkers = ["prog", "lib"]


class ConsistencyCheckerBase:
    """
    Base class for consistency checker classes.
    Checker of each type should implement check_consistency method.
    """
    def check_consistency(self, filename):
        pass


class LibChecker(ConsistencyCheckerBase):
    """
    Allows registration of a consistency checking function and verifying
    the consistency of a file.

    The function has to be in a shared library. It is then used to check
    consistency of an arbitrary file. The function has to take a file name
    as the only parameter and return an int: 0 for inconsistent, 1 for
    consistent. The prototype of the function::

        int func_name(const char* file_name)
    """

    def __init__(self, library_name, func_name):
        """
        Loads the consistency checking function from the given library.

        :param library_name: The full name of the library.
        :type library_name: str
        :param func_name: The name of the consistency
                          checking function within the library.
        :type func_name: str
        :return: None
        """
        self._lib_func = getattr(cdll.LoadLibrary(library_name), func_name)
        self._lib_func.argtypes = [c_char_p]
        self._lib_func.restype = c_int

    def check_consistency(self, filename):
        """
        Checks the consistency of a given file
        using the previously loaded function.

        :param filename: The full name of the file to be checked.
        :type filename: str
        :return: 1 if file is consistent, 0 otherwise.
        :rtype: int
        :raises: Generic exception, when no function has been loaded.
        """
        if self._lib_func is None:
            raise RuntimeError("Consistency check function not loaded")
        return self._lib_func(filename)


class ProgChecker(ConsistencyCheckerBase):
    """
    Allows registration of a consistency checking program and verifying
    the consistency of a file.
    """

    def __init__(self, bin_path, bin_args):
        self._bin_path = bin_path
        self._bin_cmd = bin_args

    def check_consistency(self, filename):
        """
        Checks the consistency of a given file
        using the previously loaded function.

        :param filename: The full name of the file to be checked.
        :type filename: str
        :return: 1 if file is consistent, 0 otherwise.
        :rtype: int
        :raises: Generic exception, when no function has been loaded.
        """
        if self._bin_path is None or self._bin_cmd is None:
            raise RuntimeError("consistency check handle not set")
        return os.system(self._bin_path + " " + self._bin_cmd + " " + filename)


def get_checker(checker_type, checker_path_args, name):

    checker_path_args = checker_path_args.split(" ", 1)
    checker_path = checker_path_args[0]

    # check for params
    if len(checker_path_args) > 1:
        args = checker_path_args[1]
    else:
        args = ""

    if not path.exists(checker_path):
        print("Invalid path:" + checker_path)
        exit(1)

    checker = None
    if checker_type == "prog":
        checker = ProgChecker(checker_path, args)
    elif checker_type == "lib":
        checker = LibChecker(checker_path, name)

    return checker



#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""Parser for user provided test configuration"""

import argparse
import os
import string
import sys
from datetime import timedelta
from collections import namedtuple

import context as ctx

try:
    import testconfig
except ImportError:
    sys.exit('Please add valid testconfig.py file - see testconfig.py.example')


def _str2list(config):
    """
    Convert the string with test sequence to equivalent list.
    example:
    _str2list("0-3,6") -->  [0, 1, 2, 3, 6]
    _str2list("1,3-5") -->  [1, 3, 4, 5]
    """
    arg = config['test_sequence']
    if not arg:
        # test_sequence not set, do nothing
        return

    seq = []
    try:
        if ',' in arg or '-' in arg:
            arg = arg.split(',')
            for number in arg:
                if '-' in number:
                    number = number.split('-')
                    begin = int(number[0])
                    end = int(number[1])
                    step = 1 if begin < end else -1
                    for x in range(begin, end + step, step):
                        seq.append(x)
                else:
                    seq.append(int(number))
        else:
            seq.append(int(arg))

    except (ValueError, IndexError):
        print('Provided test sequence "{}" is invalid'.format(arg))
        raise

    config['test_sequence'] = seq


def _str2time(config):
    """
    Convert the string with s, m, h, d suffixes to time format

    example:
    _str2time("5s")  -->  "0:00:05"
    _str2time("15m") -->  "0:15:00"
    """
    string_ = config['timeout']
    try:
        timeout = int(string_[:-1])
    except ValueError as e:
        raise ValueError("invalid timeout argument: {}".format(string_)) from e
    else:
        if "d" in string_:
            timeout = timedelta(days=timeout)
        elif "m" in string_:
            timeout = timedelta(minutes=timeout)
        elif "h" in string_:
            timeout = timedelta(hours=timeout)
        elif "s" in string_:
            timeout = timedelta(seconds=timeout)

        config['timeout'] = timeout.total_seconds()


def _str2ctx(config):
    """Convert context classes from strings to actual classes"""
    def class_from_string(name, base):
        if name == 'all':
            return base.__subclasses__()

        try:
            return next(b for b in base.__subclasses__()
                        if str(b) == name.lower())
        except StopIteration:
            print('Invalid config value: "{}".'.format(name))
            raise

    def convert_internal(key, base):
        if not isinstance(config[key], list):
            config[key] = ctx.expand(class_from_string(config[key], base))
        else:
            classes = [class_from_string(cl, base) for cl in config[key]]
            config[key] = ctx.expand(*classes)

    convert_internal('build', ctx._Build)
    convert_internal('test_type', ctx._TestType)
    convert_internal('fs', ctx._Fs)


class Configurator():
    """Parser for user test configuration"""

    def __init__(self):
        self.argparser = self._init_argparser()

    def parse_config(self):
        """
        Parse and return test execution config object. Final config is
        composed from 2 config values - values from testconfig.py file
        and values provided by command line args.
        """
        try:
            args_config = self._get_args_config()

            # The order of configs addition in 'config' initialization
            # is relevant - values from each next added config overwrite
            # values of already existing keys.
            config = {**testconfig.config, **args_config}

            self._convert_to_usable_types(config)

            # Remake dict into namedtuple for convenient fields acquisition
            return namedtuple('Config', config.keys())(**config)

        except KeyError as e:
            sys.exit("No config field '{}' found. "
                     "testconfig.py file may be invalid.".format(e.args[0]))

    def _convert_to_usable_types(self, config):
        """
        Converts config values types as parsed from user input into
        types usable by framework implementation
        """
        _str2ctx(config)
        _str2list(config)
        _str2time(config)

    def _get_args_config(self):
        """Return config values parsed from command line arguments"""

        # 'group' positional argument added only if RUNTESTS.py is the
        # execution entry point
        from_runtest = os.path.basename(sys.argv[0]) == 'RUNTESTS.py'
        if from_runtest:
            self.argparser.add_argument('group', nargs='*',
                                        help='Run only tests '
                                        'from selected groups')

        # remove possible whitespace and empty args
        sys.argv = [arg for arg in sys.argv if arg and not arg.isspace()]

        args = self.argparser.parse_args()

        if from_runtest:
            # test_sequence does not make sense if group is not set
            if args.test_sequence and not args.group:
                self.argparser.error('"--test_sequence" argument needs '
                                     'to have "group" arg set')

            # remove possible path characters added by shell hint
            args.group = [g.strip(string.punctuation) for g in args.group]

        # make into dict for type consistency
        return {k: v for k, v in vars(args).items() if v is not None}

    def _init_argparser(self):
        def ctx_choices(cls):
            return [str(c) for c in cls.__subclasses__()]

        parser = argparse.ArgumentParser()
        parser.add_argument('--fs_dir_force_pmem', type=int,
                            help='set PMEM_IS_PMEM_FORCE for tests run on'
                            ' pmem fs')
        parser.add_argument('-l', '--unittest_log_level', type=int,
                            help='set log level. 0 - silent, 1 - normal, '
                            '2 - verbose')
        parser.add_argument('--keep_going', type=bool,
                            help='continue execution despite test fails')
        parser.add_argument('-b', dest='build',
                            help='run only specified build type',
                            choices=ctx_choices(ctx._Build), nargs='*')
        parser.add_argument('-f', dest='fs',
                            choices=ctx_choices(ctx._Fs), nargs='*',
                            help='run tests on specified filesystem types.')
        parser.add_argument('-t', dest='test_type',
                            help='run only specified test type where'
                            'check = short + medium',
                            choices=ctx_choices(ctx._TestType), nargs='*')
        parser.add_argument('-o', dest='timeout',
                            help="set timeout for test execution timeout:"
                            "integer with an optional suffix:''s' for seconds,"
                            "'m' for minutes, 'h' for hours or 'd' for days.")
        parser.add_argument('-u', dest='test_sequence',
                            help='run only tests from specified test sequence '
                            'e.g.: 0-2,5 will execute TEST0, '
                            'TEST1, TEST2 and TEST5',
                            default='')
        return parser



#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""Base tests class and its functionalities"""

import builtins
import itertools
import shutil
import subprocess as sp
import sys
from datetime import datetime
from os import listdir, makedirs, path

import context as ctx
import helpers as hlp


if not hasattr(builtins, 'testcases'):
    builtins.testcases = []


def get_testcases():
    """"Get list of testcases imported from src/test tree"""
    return builtins.testcases


# test case attributes that refer to selected context classes, their
# respective config field names and context base classes
CTX_COMPONENTS = (
    ('build', ctx._Build),
    ('fs', ctx._Fs)
)


class Fail(Exception):
    """Thrown when test fails"""
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class Any:
    """
    Test context attribute signifying that specific context value is not
    relevant for the test outcome and it should be run only once in some
    viable context
    """
    @classmethod
    def get(cls, conf_ctx):
        """Get specific context value to be run"""
        for c in conf_ctx:
            if c.is_preferred:
                # pick preferred if found
                return c
        # if no preferred is found, pick the first one
        return conf_ctx[0]


class _TestCase(type):
    """Metaclass for BaseTest that is used for registering imported tests"""

    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        # only classes deriving from BaseTest are meant to be used
        if cls.__base__.__name__ != 'BaseTest':
            return

        # globally register class as test case
        builtins.testcases.append(cls)

        # expand values of context classes attributes
        for attr, _ in CTX_COMPONENTS:
            if hasattr(cls, attr):
                val = getattr(cls, attr)
                if isinstance(val, list):
                    val = ctx.expand(*val)
                setattr(cls, attr, val)

        cls.name = cls.__name__
        try:
            cls.testnum = int(cls.__name__.replace('TEST', ''))
        except ValueError as e:
            print('Invalid test class name {}, should be "TEST[number]"'
                  .format(cls.__name__))
            raise e


class BaseTest(metaclass=_TestCase):
    """Every test case need to inherit from this class"""
    test_type = ctx.Medium
    match = True

    def __repr__(self):
        return '{}/{}'.format(self.group, self.__class__.__name__)

    def __init__(self, config):
        if self.__module__ == '__main__':
            self.cwd = path.dirname(path.abspath(sys.argv[0]))
        else:
            self.cwd = self.__module__

        self.config = config
        self.msg = hlp.Message(config)
        self.group = path.basename(self.cwd)

        self.testdir = self.group + '_' + str(self.testnum)
        self.utenv = self._get_utenv()
        self.ctxs = self._get_contexts()

    def _ctx_attrs_init(self):
        """
        Initialize test class attributes referring to selected context
        parameters (like build, fs). If attribute was not set by subclassing
        test, its value is set to respective config value. If it was set,
        it is filtered through config values.
        """

        for attr, base in CTX_COMPONENTS:
            conf_val = getattr(self.config, attr)
            if hasattr(self, attr):
                test_val = getattr(self, attr)
                if test_val == Any:
                    ctx_val = Any.get(conf_val)
                else:
                    ctx_val = hlp.filter_contexts(conf_val, test_val)
            else:
                ctx_val = hlp.filter_contexts(conf_val, None)
            setattr(self, attr, base.factory(self.config, *ctx_val))

    def _get_contexts(self):
        """Initialize context classes used for each test run"""
        self._ctx_attrs_init()

        ctxs = []
        if self.test_type not in self.config.test_type:
            return ctxs

        # generate combination sets of context components
        ctx_sets = itertools.product(self.fs, self.build)
        for cs in ctx_sets:
            ctxs.append(ctx.Context(self, self.config, fs=cs[0], build=cs[1]))

        return ctxs

    def _get_utenv(self):
        """Get environment variables values used by C test framework"""
        return {
            'UNITTEST_NAME': str(self),
            'UNITTEST_LOG_LEVEL': str(self.config.unittest_log_level),
            'UNITTEST_NUM': str(self.testnum)
        }

    def _execute(self):
        """
        Execute test for each context, return 1 if at least
        one test failed
        """
        failed = False
        for c in self.ctxs:
            self.msg.print('{}: SETUP\t({}/{}/{})'
                           .format(self, self.test_type, c.fs, c.build))

            self.setup(c)
            start_time = datetime.now()
            try:
                self.run(c)
                self.elapsed = (datetime.now() - start_time).total_seconds()
                self.check()

            except Fail as f:
                failed = True
                print(f)
                print('{}: {}FAILED{}\t({}/{}/{})'
                      .format(self, hlp.Color.RED, hlp.Color.END,
                              self.test_type, c.fs, c.build))
                if not self.config.keep_going:
                    sys.exit(1)

            except sp.TimeoutExpired:
                failed = True
                print('{}: {}TIMEOUT{}\t({}/{}/{})'
                      .format(self, hlp.Color.RED, hlp.Color.END,
                              self.test_type, c.fs, c.build))
                if not self.config.keep_going:
                    sys.exit(1)

            else:
                self._test_passed()

            finally:
                self.clean(c)

        if failed:
            return 1
        return 0

    def setup(self, ctx):
        """Test setup"""
        if not path.exists(ctx.testdir):
            makedirs(ctx.testdir)

    def run(self, ctx):
        """
        Main test body, run with specific context provided through
        Context class instance. Needs to be implemented by each test
        """
        raise NotImplementedError('{} does not implement run() method'.format(
            self.__class__))

    def check(self):
        """Run additional test checks"""
        if self.match:
            self._run_match()

    def fail(self, msg):
        raise Fail(msg)

    def _run_match(self):
        """Match log files"""
        cwd_listdir = [path.join(self.cwd, f) for f in listdir(self.cwd)]

        suffix = '{}.log.match'.format(self.testnum)

        def is_matchfile(f):
            """Match file ends with specific suffix and a char before suffix
            is not a digit"""
            before_suffix = -len(suffix) - 1
            return path.isfile(f) and f.endswith(suffix) and \
                not f[before_suffix].isdigit()

        match_files = filter(is_matchfile, cwd_listdir)
        prefix = 'perl ' if sys.platform == 'win32' else ''
        match_cmd = prefix + path.join(hlp.ROOTDIR, 'match')

        for mf in match_files:
            cmd = '{} {}'.format(match_cmd, mf)
            proc = sp.run(cmd.split(), stdout=sp.PIPE, cwd=self.cwd,
                          stderr=sp.STDOUT, universal_newlines=True)
            if proc.returncode != 0:
                self.fail(proc.stdout)
            else:
                self.msg.print_verbose(proc.stdout)

    def clean(self, ctx):
        """Remove directory, even if it is not empty"""
        shutil.rmtree(ctx.testdir, ignore_errors=True)

    def _test_passed(self):
        """Print message specific for passed test"""
        if self.config.tm:
            tm = '\t\t\t[{:06.3F} s]'.format(self.elapsed)
        else:
            tm = ''

        self.msg.print('{}: {}PASS{} {}'
                       .format(self, hlp.Color.GREEN, hlp.Color.END, tm))



#!/usr/bin/env python3
#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
"""
Interpreter managing test group specific TESTS.py file execution.
It parses test classes from interpreted file, handles command line arguments
and executes tests using provided configuration.
"""


import importlib.util as importutil
import os
import sys

from testframework import BaseTest, Configurator, run_tests_common


def run_testcases():
    """Parse user configuration, run test cases"""
    config = Configurator().parse_config()
    testcases = BaseTest.__subclasses__()
    return run_tests_common(testcases, config)


def main():
    # Interpreter receives TESTS.py file as first argument
    if len(sys.argv) < 2:
        sys.exit('Provide test file to run')
    testfile = sys.argv[1]

    # Remove TESTS.py file from args, the rest of the args is parsed as a
    # test configuration
    sys.argv.pop(1)

    # import TESTS.py as a module
    testfile_dir = os.path.abspath(os.path.dirname(testfile))
    spec = importutil.spec_from_file_location(testfile_dir, testfile)
    module = importutil.module_from_spec(spec)
    spec.loader.exec_module(module)

    sys.exit(run_testcases())


if __name__ == '__main__':
    main()



# Copyright 2018-2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import memoryoperations as memops
import reorderengines
from reorderexceptions import InconsistentFileException
from reorderexceptions import NotSupportedOperationException


class State:
    """
    The base class of all states.

    :ivar _context: The reordering context.
    :type _context: opscontext.OpsContext
    :ivar trans_stores: The list of unflushed stores.
    :type trans_stores: list of :class:`memoryoperations.Store`
    """
    trans_stores = []

    def __init__(self, context):
        """
        Default state constructor.

        :param context: The context of the reordering.
        :type context: opscontext.OpsContext
        """
        self._context = context

    def next(self, in_op):
        """
        Go to the next state based on the input.

        :Note:
            The next state might in fact be the same state.

        :param in_op: The state switch trigger operation.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: The next state.
        :rtype: subclass of :class:`State`
        """
        raise NotImplementedError

    def run(self, in_op):
        """
        Perform the required operation in this state.

        :param in_op: The operation to be performed in this state.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: None
        """
        raise NotImplementedError


class InitState(State):
    """
    The initial no-op state.
    """
    def __init__(self, context):
        """
        Saves the reordering context.

        :param context: The reordering context.
        :type context: opscontext.OpsContext
        """
        super(InitState, self).__init__(context)

    def next(self, in_op):
        """
        Switch to the next valid state.

        :param in_op: Ignored.
        :return: The next valid state.
        :rtype: CollectingState
        """
        return CollectingState(self._context)

    def run(self, in_op):
        """
        Does nothing.

        :param in_op: Ignored.
        :return: always True
        """
        return True


class CollectingState(State):
    """
    Collects appropriate operations.

    This state mostly aggregates stores and flushes. It also
    validates which stores will be made persistent and passes
    them on to the next state.

    :ivar _ops_list: The list of collected stores.
    :type _ops_list: list of :class:`memoryoperations.Store`
    :ivar _inner_state: The internal state of operations.
    :type _inner_state: str
    """
    def __init__(self, context):
        """
        Saves the reordering context.

        :param context: The reordering context.
        :type context: opscontext.OpsContext
        """
        super(CollectingState, self).__init__(context)
        self._ops_list = []
        self._ops_list += State.trans_stores
        self._inner_state = "init"

    def next(self, in_op):
        """
        Switch to the next valid state.

        :param in_op: The state switch trigger operation.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: The next state.
        :rtype: subclass of :class:`State`
        """
        if isinstance(in_op, memops.Fence) and \
                self._inner_state is "flush":
            return ReplayingState(self._ops_list, self._context)
        else:
            return self

    def run(self, in_op):
        """
        Perform operations in this state.

        Based on the type of operation, different handling is employed.
        The recognized and handled types of operations are:

            * :class:`memoryoperations.ReorderBase`
            * :class:`memoryoperations.FlushBase`
            * :class:`memoryoperations.Store`
            * :class:`memoryoperations.Register_file`

        :param in_op: The operation to be performed in this state.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: always True
        """
        self.move_inner_state(in_op)
        if isinstance(in_op, memops.ReorderBase):
            self.substitute_reorder(in_op)
        elif isinstance(in_op, memops.FlushBase):
            self.flush_stores(in_op)
        elif isinstance(in_op, memops.Store):
            self._ops_list.append(in_op)
        elif isinstance(in_op, memops.Register_file):
            self.reg_file(in_op)

        return True

    def substitute_reorder(self, order_ops):
        """
        Changes the reordering engine based on the log marker class.

        :param order_ops: The reordering marker class.
        :type order_ops: subclass of :class:`memoryoperations.ReorderBase`
        :return: None
        """
        if isinstance(order_ops, memops.ReorderFull):
            self._context.reorder_engine = \
                reorderengines.FullReorderEngine()
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.ReorderPartial):
            # TODO add macro in valgrind or
            # parameter inside the tool to support parameters?
            self._context.reorder_engine = \
                 reorderengines.RandomPartialReorderEngine(3)
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.ReorderAccumulative):
            self._context.reorder_engine = \
                reorderengines.AccumulativeReorderEngine()
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.ReorderReverseAccumulative):
            self._context.reorder_engine = \
                reorderengines.AccumulativeReverseReorderEngine()
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.NoReorderDoCheck):
            self._context.reorder_engine = reorderengines.NoReorderEngine()
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.NoReorderNoCheck):
            self._context.reorder_engine = reorderengines.NoCheckerEngine()
            self._context.test_on_barrier = \
                self._context.reorder_engine.test_on_barrier
        elif isinstance(order_ops, memops.ReorderDefault):
            self._context.reorder_engine = self._context.default_engine
            self._context.test_on_barrier = self._context.default_barrier
        else:
            raise NotSupportedOperationException(
                                   "Not supported reorder engine: {}"
                                   .format(order_ops))

    def flush_stores(self, flush_op):
        """
        Marks appropriate stores as flushed.

        Does not align the flush, the log is expected to have the
        flushes properly aligned.

        :param flush_op: The flush operation marker.
        :type flush_op: subclass of :class:`memoryoperations.FlushBase`
        :return: None
        """
        for st in self._ops_list:
            if flush_op.is_in_flush(st):
                st.flushed = True

    def reg_file(self, file_op):
        """
        Register a new file mapped into virtual memory.

        :param file_op: File registration operation marker.
        :type file_op: memoryoperations.Register_file
        :return: None
        """
        self._context.file_handler.add_file(file_op.name,
                                            file_op.address,
                                            file_op.size)

    def move_inner_state(self, in_op):
        """
        Tracks the internal state of the collection.

        The collected stores need to be processed only at specific moments -
        after full persistent memory barriers (flush-fence).

        :param in_op: The performed operation.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: None
        """
        if isinstance(in_op, memops.Store) and \
                self._inner_state is "init":
            self._inner_state = "dirty"
        elif isinstance(in_op, memops.FlushBase) and \
                self._inner_state is "dirty":
            self._inner_state = "flush"
        elif isinstance(in_op, memops.Fence) and \
                self._inner_state is "flush":
            self._inner_state = "fence"
        elif isinstance(in_op, memops.Flush) and \
                self._inner_state is "init":
            self._inner_state = "flush"


class ReplayingState(State):
    """
    Replays all collected stores according to the reordering context.

    :ivar _ops_list: The list of stores to be reordered and replayed.
    :type _ops_list: list of :class:`memoryoperations.Store`
    """
    def __init__(self, in_ops_list, context):
        """

        :param in_ops_list:
        :param context:
        :return:
        """
        super(ReplayingState, self).__init__(context)
        self._ops_list = in_ops_list

    def next(self, in_op):
        """
        Switches to the collecting state regardless of the input.

        :param in_op: Ignored.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: The next state.
        :rtype: CollectingState
        """
        return CollectingState(self._context)

    def run(self, in_op):
        """
        Perform operations in this state.

        The replaying state performs reordering and if necessary checks
        the consistency of the registered files. The decisions and
        type of reordering to be used is defined by the context.

        :param in_op: The operation to be performed in this state.
        :type in_op: subclass of :class:`memoryoperations.BaseOperation`
        :return: State of consistency check.
        """
        # specifies consistency state of sequence
        consistency = True

        # consider only flushed stores
        flushed_stores = list(filter(lambda x: x.flushed, self._ops_list))

        # already flushed stores should be removed from the transitive list
        common_part = list(set(flushed_stores) & set(State.trans_stores))
        State.trans_stores = list(set(State.trans_stores) - set(common_part))

        # do not add redundant stores
        State.trans_stores = list(set(State.trans_stores) |
                                  set(list(filter(lambda x: x.flushed is False,
                                      self._ops_list))))

        if self._context.test_on_barrier:
            for seq in self._context.reorder_engine.generate_sequence(
                                                              flushed_stores):
                for op in seq:
                    # do stores
                    self._context.file_handler.do_store(op)
                # check consistency of all files
                try:
                    self._context.file_handler.check_consistency()
                except InconsistentFileException as e:
                    consistency = False
                    self._context.logger.warning(e)
                    stacktrace = "Call trace:\n"
                    for num, op in enumerate(seq):
                        stacktrace += "Store [{}]:\n".format(num)
                        stacktrace += str(op.trace)
                    self._context.logger.warning(stacktrace)

                for op in reversed(seq):
                    # revert the changes
                    self._context.file_handler.do_revert(op)
        # write all flushed stores
        for op in flushed_stores:
            self._context.file_handler.do_store(op)

        return consistency


class StateMachine:
    """
    The state machine driver.

    :ivar _curr_state: The current state.
    :type _curr_state: subclass of :class:`State`
    """
    def __init__(self, init_state):
        """
        Initialize the state machine with a specified state.

        :param init_state: The initial state to be used.
        :type init_state: subclass of :class:`State`
        """
        self._curr_state = init_state

    def run_all(self, operations):
        """
        Starts the state machine.

        :param operations: The operations to be performed by the state
            machine.
        :type operations: list of :class:`memoryoperations.BaseOperation`
        :return: None
        """
        all_consistent = True
        for ops in operations:
            self._curr_state = self._curr_state.next(ops)
            check = self._curr_state.run(ops)
            if check is False:
                all_consistent = check

        return all_consistent



#!/usr/bin/env python3
#
# Copyright 2019, Intel Corporation
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""Main script for unit tests execution"""

import sys
import os
from os import path
sys.path.insert(1, path.abspath(path.join(path.dirname(__file__), 'unittest')))

# flake8 issues silenced:
# E402 - import statements not at the top of the file because of adding
# directory to path

import importlib.util as importutil  # noqa E402

import helpers  # noqa E402
from basetest import get_testcases  # noqa E402
from configurator import Configurator  # noqa E402


def _import_testfiles():
    """
    Traverse through "src/test" directory, find all "TESTS.py" files and
    import them as modules. Set imported module name to file directory path.
    """
    for root, _, files in os.walk(helpers.ROOTDIR):
        for name in files:
            if name == 'TESTS.py':
                testfile = path.join(root, name)
                spec = importutil.spec_from_file_location(
                    path.dirname(testfile), testfile)
                module = importutil.module_from_spec(spec)
                spec.loader.exec_module(module)


def _run_tests(config):
    """Run tests imported from repository tree from selected groups"""
    testcases = get_testcases()

    if config.group:
        # filter selected groups
        testcases = [t for t in testcases
                     if path.basename(t.__module__) in config.group]

        if not testcases:
            sys.exit('No testcases found for selected groups. '
                     'Provided group name may be invalid.')

    return helpers.run_tests_common(testcases, config)


def main():
    config = Configurator().parse_config()
    _import_testfiles()
    sys.exit(_run_tests(config))


if __name__ == "__main__":
    main()
